// No copyright
#pragma once

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>


#define ACG_CHUNK_PAD   0x00U
#define ACG_CHUNK_BLOB  0x01U
#define ACG_CHUNK_CTX   0x02U
#define ACG_CHUNK_LOC   0x03U
#define ACG_CHUNK_BEGIN 0x04U
#define ACG_CHUNK_END   0x05U
#define ACG_CHUNK_CHECK 0x06U

#define ACG_UNITBITS_CHUNK_ID 4U


static inline uint64_t acg_util_fullbits(uint8_t bits) {
  return (((uint64_t) 1U) << bits) - 1U;
}
static inline uint64_t acg_util_umin(uint64_t a, uint64_t b) {
  return a > b? b: a;
}


struct acg_stream {
  uint8_t* buffer;
  uint64_t size_bytes;

  uint64_t cursor_bits;
};

static inline bool acg_stream_write_pad(struct acg_stream* s) {
  assert(NULL != s);

  s->cursor_bits = (s->cursor_bits+7U)/8U*8U;
  return true;
}
static inline bool acg_stream_write_bool(struct acg_stream* s, bool v) {
  assert(NULL != s);

  if (s->size_bytes <= s->cursor_bits) {
    return false;
  }
  uint64_t* const c = &s->cursor_bits;
  s->buffer[*c/8U] |= (!!v) << (*c%8U);
  ++*c;
  return true;
}
static inline bool acg_stream_write_uint(struct acg_stream* s, uint64_t v, uint8_t unit_bits) {
  assert(NULL != s);
  assert(2U  <= unit_bits);
  assert(65U >= unit_bits);

  uint8_t*  const buf = s->buffer;
  uint64_t* const c   = &s->cursor_bits;

  const uint8_t unit_data_bits = unit_bits - 1U;
  while (v) {
    if (s->size_bytes*8U - *c < unit_bits) {
      return false;
    }

    const uint64_t next_v = v >> unit_data_bits;
    acg_stream_write_bool(s, !!next_v);

    const uint8_t frag_bits = acg_util_umin(unit_data_bits, (8U - *c%8U)%8U);
    if (frag_bits > 0U) {
      buf[*c/8U] |= (v & acg_util_fullbits(frag_bits)) << (*c%8U);
      *c += frag_bits;
    }

    const uint8_t  remain_bits = unit_data_bits - frag_bits;
    const uint64_t v_part      = (v >> frag_bits) & acg_util_fullbits(remain_bits);
    memcpy(&buf[*c/8U], &v_part, (remain_bits + 7U) / 8U);
    *c += remain_bits;

    v = next_v;
  }
  return true;
}
static inline bool acg_stream_write_int(struct acg_stream* s, int64_t v, uint8_t unit_bits) {
  assert(NULL != s);
  assert(2U  <= unit_bits);
  assert(65U >= unit_bits);

  if (v >= 0U) {
    return
        acg_stream_write_bool(s, false) &&
        acg_stream_write_uint(s, (uint64_t) v, unit_bits);
  } else {
    return
        acg_stream_write_bool(s, true) &&
        acg_stream_write_uint(s, (uint64_t) -v, unit_bits);
  }
}

static inline bool acg_stream_write_chunk_pad(struct acg_stream* s) {
  assert(NULL != s);

  if (s->cursor_bits%8U == 0U) {
    return true;
  }
  if (s->size_bytes <= s->cursor_bits/8U) {
    return false;
  }
  return
      acg_stream_write_uint(s, ACG_CHUNK_PAD, ACG_UNITBITS_CHUNK_ID) &&
      acg_stream_write_pad(s);
}
static inline bool acg_stream_write_chunk_blob(struct acg_stream* s, uint64_t size_bytes, const uint8_t* buf) {
  assert(NULL != s);

  if (!acg_stream_write_uint(s, ACG_CHUNK_BLOB, ACG_UNITBITS_CHUNK_ID)) {
    return false;
  }
  if (!acg_stream_write_uint(s, size_bytes, 8U)) {
    return false;
  }
  if (!acg_stream_write_pad(s)) {
    return false;
  }

  uint64_t* const c = &s->cursor_bits;

  const uint64_t avail_bytes = s->size_bytes - *c/8U;
  if (avail_bytes < size_bytes) {
    return false;
  }
  memcpy(&s->buffer[*c/8], buf, size_bytes);

  const uint64_t size_bits = size_bytes * 8U;
  *c += size_bits;
  return true;
}
static inline bool acg_stream_write_chunk_begin(struct acg_stream* s) {
  assert(NULL != s);

  if (!acg_stream_write_uint(s, ACG_CHUNK_BEGIN, ACG_UNITBITS_CHUNK_ID)) {
    return false;
  }
  return true;
}
static inline bool acg_stream_write_chunk_end(struct acg_stream* s) {
  assert(NULL != s);

  if (!acg_stream_write_uint(s, ACG_CHUNK_END, ACG_UNITBITS_CHUNK_ID)) {
    return false;
  }
  return true;
}
static inline bool acg_stream_write_chunk_check(struct acg_stream* s) {
  assert(NULL != s);

  if (!acg_stream_write_uint(s, ACG_CHUNK_CHECK, ACG_UNITBITS_CHUNK_ID)) {
    return false;
  }
  return true;
}
